
class DictObject(dict):
    '''
    Provides a JavaScript-esque Object where attribute access is
    equivalent to dict access.

    >>> a = DictObject()
    >>> a['a'] = 1
    >>> a.a
    1
    >>> a.a = 3
    >>> a['a']
    3
    '''
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError('%s object has no attribute %s' % (self, key))

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]



from collections import namedtuple
from operator import attrgetter


class Ripper(object):
    '''
    Extremely light-weight serialiser for very simple cases.

    See http://musings.tinbrain.net/blog/2015/aug/14/serialiser-hurry/ for
    details.
    '''
    def __init__(self, *args, **kwargs):
        for arg in args:
            kwargs.setdefault(arg, arg)
        self.getter = attrgetter(*kwargs.values())
        self.tup = namedtuple('tup', kwargs.keys())

    def __call__(self, obj):
        return self.tup._make(self.getter(obj))._asdict()




from collections import defaultdict
from inspect import classify_class_attrs

from django.core.exceptions import ValidationError
from django.db.models.fields import NOT_PROVIDED
from django.utils.six import with_metaclass

from .fields import field
from .utils import DictObject


class MetaMapper(type):

    def __new__(mcs, name, bases, attrs):
        new_class = super(MetaMapper, mcs).__new__(mcs, name, bases, attrs)
        fields = {
            name: prop
            for name, kind, cls, prop in classify_class_attrs(new_class)
            if isinstance(prop, field)
        }

        new_class._fields = fields
        new_class._field_names = tuple(fields.keys())

        return new_class


class DataMapper(with_metaclass(MetaMapper)):
    '''
    DataMapper class.

    Provides a proxy class for retrieving and updating attributes on your objects.
    '''
    def __init__(self, obj=None, **kwargs):
        '''
        :param: obj Optionally bind to object
        :param: **kwargs Extra context (stored as self._context)
        '''

        if obj is None:
            obj = DictObject()
        self._obj = obj
        self._context = kwargs

    def __lshift__(self, other):
        '''
        Allow implicit reduction using:

        >>> data = mapper << obj
        '''
        self._obj = other
        return self._reduce()

    def __rlshift__(self, other):
        '''
        Allow implicit patch using:

        >>> obj = data >> mapper
        '''
        return self._patch(other)

    def _reduce(self):
        '''
        Reduce our instance to its serialisable state.

        Returns a dict.
        '''
        return {
            name: getattr(self, name)
            for name in self._field_names
        }

    def _clean(self, data, full=True):
        '''
        Hook for finall pass validation.

        full indicates if this is an _apply (True) or _patch (False)
        Should update self._errors dict.
        '''
        return

    def _patch(self, data):
        '''
        Update an instance from supplied data.
        '''

        errors = defaultdict(list)

        for name in self._field_names:
            try:
                value = data[name]
            except KeyError:
                continue
            try:
                setattr(self, name, value)
            except ValidationError as e:
                errors[name].append(e)

        self._errors = dict(errors)

        # Allow a final pass of cleaning
        self._clean(data, full=False)

        if self._errors:
            raise ValidationError(self._errors)

        return self._obj

    def _apply(self, data):
        '''
        Update an instance from supplied data.

        All fields marked required=True MUST be provided.
        All fields omitted will have their default used, if provided.
        '''
        errors = defaultdict(list)

        for name in self._field_names:
            required = getattr(self._fields[name], 'required', True)
            default = getattr(self._fields[name], 'default', NOT_PROVIDED)

            try:
                value = data[name]
            except KeyError:
                if required:
                    if default is NOT_PROVIDED:
                        errors[name].append(
                            ValidationError('This field is required')
                        )
                        continue
                elif default is NOT_PROVIDED:
                    continue
                value = default
                if callable(value):
                    value = value()
            try:
                setattr(self, name, value)
            except ValidationError as e:
                errors[name].append(e)

        self._errors = dict(errors)

        # Allow a final pass of cleaning
        self._clean(data, full=True)

        if self._errors:
            raise ValidationError(self._errors)

        return self._obj




from django import VERSION
from django.core.exceptions import ValidationError
from django.db.models.fields import NOT_PROVIDED
from django.utils.six import with_metaclass

from . import filters
from .fields import Field
from .mappers import DataMapper, MetaMapper as BaseMetaMapper

# Map of ModelField name -> list of filters
FIELD_FILTERS = {
    'DateField': [filters.DateFilter],
    'TimeField': [filters.TimeFilter],
    'DateTimeField': [filters.DateTimeFilter],

    'BooleanField': [filters.BooleanFilter],
    'IntegerField': [filters.IntegerFilter],
}


if VERSION < (1, 8):
    def is_relfield(field):
        return field.rel.to if field.rel else None
else:
    def is_relfield(field):
        return field.related_model if field.is_relation else None


class Options(object):
    def __init__(self, meta):
        self.model = getattr(meta, 'model', None)
        fields = getattr(meta, 'fields', [])
        if fields != '__all__':
            fields = list(fields)
        self.fields = fields
        self.exclude = list(getattr(meta, 'exclude', []))
        self.required = getattr(meta, 'required', {})


class MetaMapper(BaseMetaMapper):

    def __new__(mcs, name, bases, attrs):
        meta = Options(attrs.get('Meta', None))

        if meta.model is None:
            if name not in ['NewBase', 'ModelDataMapper']:
                raise ValueError('model not defined on class Meta')
        else:
            # XXX Does the top base have all fields?

            for model_field in meta.model._meta.fields:
                if model_field.name in attrs:
                    continue
                if model_field.name in meta.exclude:
                    continue
                if meta.fields != '__all__' and \
                   model_field.name not in meta.fields:
                    # Ensure model validation is told to exclude this
                    meta.exclude.append(model_field.name)
                    continue

                # XXX Magic for field types
                kwargs = {}
                if model_field.null is True and model_field.default is NOT_PROVIDED:
                    kwargs['default'] = None
                else:
                    kwargs['default'] = model_field.default
                kwargs['required'] = meta.required.get(
                    model_field.name,
                    not model_field.blank,
                )
                kwargs['filters'] = FIELD_FILTERS.get(
                    model_field.__class__.__name__,
                    []
                )[:]
                if not model_field.null:
                    kwargs['filters'].insert(0, filters.NotNullFilter)
                related_model = is_relfield(model_field)
                if related_model:
                    kwargs['filters'].insert(0, ModelFilter(related_model))
                attrs[model_field.name] = Field(model_field.name, **kwargs)

        attrs['_meta'] = meta

        return super(MetaMapper, mcs).__new__(mcs, name, bases, attrs)


class ModelDataMapper(with_metaclass(MetaMapper, DataMapper)):

    def __init__(self, obj=None, **kwargs):
        if obj is None:
            obj = self._meta.model()
        super(ModelDataMapper, self).__init__(obj, **kwargs)

    def __irshift__(self, other):
        '''
        Allow implicit "update new instance" using:

        >>> obj = data >>= mapper
        '''
        self.obj = self._meta.model()
        return self._apply(other)

    def _clean(self, data, full=True):
        try:
            self._obj.full_clean(exclude=self._meta.exclude)
        except ValidationError as e:
            for k, v in e.message_dict.items():
                self._errors.setdefault(k, []).extend(v)


class ModelFilter(filters.Filter):
    def __init__(self, model=None, queryset=None):
        if model is None and queryset is None:
            raise ValueError('Must specify either "model" or "queryset".')
        if queryset:
            self.queryset = queryset
        else:
            self.queryset = model._default_manager.all()

    def to_python(self, value):
        try:
            return self.queryset.get(pk=value)
        except self.queryset.model.DoesNotExist:
            raise ValidationError('Not a valid pk for {}: {}'.format(self.queryset.model._meta.verbose_name, value))

    def from_python(self, value):
        return value.pk



import datetime

from django.core.exceptions import ValidationError


class Filter(object):
    '''
    Provides casting and validation functions for Fields.
    '''
    @staticmethod
    def to_python(value):
        return value

    @staticmethod
    def from_python(value):
        return value


class NotNullFilter(Filter):
    @staticmethod
    def to_python(value):
        if value is None:
            raise ValidationError('May not be null')
        return value


class BooleanFilter(Filter):
    @staticmethod
    def to_python(value):
        if value is None:
            return value
        if isinstance(value, bool):
            return value
        return value.lower() in (1, '1', 't', 'y', 'true')


class _CastFilter(Filter):
    @classmethod
    def to_python(self, value):
        if value is None or isinstance(value, self.type_class):
            return value
        try:
            return self.type_class(value)
        except ValueError as e:
            raise ValidationError(e.args[0])


class IntegerFilter(_CastFilter):
    type_class = int


class FloatFilter(_CastFilter):
    type_class = float


class TimeFilter(object):
    @staticmethod
    def to_python(value):
        if value is None or isinstance(value, datetime.time):
            return value
        return datetime.datetime.strptime(value, '%H:%M:%S').time()

    @staticmethod
    def from_python(value):
        if value is None:
            return value
        return value.replace(microsecond=0).isoformat()


class DateFilter(object):
    @staticmethod
    def to_python(value):
        if value is None or isinstance(value, datetime.date):
            return value
        return datetime.datetime.strptime(value, '%Y-%m-%d').date()

    @staticmethod
    def from_python(value):
        if value is None:
            return value
        return value.isoformat()


class DateTimeFilter(object):
    @staticmethod
    def to_python(value):
        if value is None or isinstance(value, datetime.datetime):
            return value
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')

    @staticmethod
    def from_python(value):
        if value is None:
            return value
        return value.replace(microsecond=0).isoformat(' ')




from django.http import StreamingHttpResponse

from .simplecsv import Writer


class ExportCsv(object):
    '''
    A factory class for Admin Actions to export a model as CSV.

    actions = [ ExportCsv(mymapper, 'Export as CSV'), ]

    '''
    def __init__(self, mapper=None, label=None, **opts):
        self.mapper = mapper
        self.opts = opts
        if label:
            self.short_description = label
        self.__name__ = label or 'ExportCsv'

    def __call__(self, admin, request, queryset):
        mapper = self.mapper()

        select_related = self.opts.get('select_related', None)
        if select_related:
            queryset = queryset.select_related(*select_related)
        prefetch_related = self.opts.get('prefetch_related', None)
        if prefetch_related:
            queryset = queryset.prefetch_related(*prefetch_related)

        def inner():
            csv = Writer(fields=self.opts.get('fields', mapper._fields.keys()))
            yield csv.write_headers()
            for obj in queryset:
                data = mapper << obj
                yield csv.write_dict(data)

        response = StreamingHttpResponse(inner(), content_type='text/csv')
        filename = self.opts.get('filename', 'export_{classname}.csv')
        if callable(filename):
            filename = filename(admin)
        else:
            filename = filename.format(
                classname=admin.__class__.__name__,
                model=admin.model._meta.module_name,
                app_label=admin.model._meta.app_label,
            )
        response['Content-Disposition'] = 'attachment; filename=%s' % filename

        return response




import inspect
import json

from django.views.generic import View

from nap import http
from nap.utils import JsonMixin

RPC_MARKER = '_rpc'


def method(view):
    '''Mark a view as accessible via RPC'''
    setattr(view, RPC_MARKER, True)
    return view


def is_rpc_method(m):
    '''Helper for checking if something is marked as a pubishable method.'''
    return getattr(m, RPC_MARKER, False)


class RPCMixin(JsonMixin):
    '''Mix in to a standard View to provide RPC actions'''
    permit_introspect = False

    def dispatch(self, request, *args, **kwargs):
        '''
        Check if this is a POST and has the X-RPC-Action header.

        If so, dispatch the request to the method, if it exists and is
        publishable.

        Otherwise behaves as normal.
        '''
        method = request.META.get('HTTP_X_RPC_ACTION', None)
        if request.method != 'POST' or method is None:
            return super(RPCMixin, self).dispatch(request, *args, **kwargs)

        func = getattr(self, method, None)
        if not is_rpc_method(func):
            return http.PreconditionFailed()

        try:
            data = self.get_request_data({})
            # Ensure data is valid for passing as **kwargs
            (lambda **kwargs: None)(**data)
        except (ValueError, TypeError):
            return http.BadRequest()

        resp = self.execute(func, data)

        return http.JsonResponse(resp)

    def options(self, request, *args, **kwargs):
        '''
        Default OPTIONS handler.

        Returns introspection data.
        '''
        response = super(RPCMixin, self).options(request, *args, **kwargs)
        if self.permit_introspect:
            response['Content-Type'] = 'application/json'
            response.write(json.dumps(self._introspect()))
        return response

    def _introspect(self):
        '''
        Provides a list of methods available on this view.
        '''
        methods = {}
        for name, prop in inspect.getmembers(self, is_rpc_method):
            argspec = inspect.getargspec(prop)
            methods[name] = {
                'args': argspec.args[1:],
                'doc': inspect.getdoc(prop),
                'defaults': argspec.defaults,
            }
        return methods

    def execute(self, handler, data):
        '''Helpful hook to ease wrapping the handler'''
        return handler(**data)


class RPCView(RPCMixin, View):
    '''Courtesy class to avoid having to mix it yourself.'''
    pass



from setuptools import setup, find_packages

setup(
    name='django-nap',
    version='0.20.2',
    description='A light REST tool for Django',
    author='Curtis Maloney',
    author_email='curtis@tinbrain.net',
    url='http://github.com/funkybob/django-nap',
    keywords=['django', 'json', 'rest'],
    packages = find_packages(exclude=('tests*',)),
    zip_safe=False,
    classifiers = [
        'Environment :: Web Environment',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    requires = [
        'Django (>=1.7)',
    ],
    install_requires = [
        'Django>=1.7',
    ],
)
